#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Cloud Alloy DB operators."""

from __future__ import annotations

from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any

from google.api_core.exceptions import NotFound
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud import alloydb_v1

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.alloy_db import AlloyDbHook
from airflow.providers.google.cloud.links.alloy_db import (
    AlloyDBBackupsLink,
    AlloyDBClusterLink,
    AlloyDBUsersLink,
)
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

if TYPE_CHECKING:
    import proto
    from google.api_core.operation import Operation
    from google.api_core.retry import Retry
    from google.protobuf.field_mask_pb2 import FieldMask

    from airflow.providers.common.compat.sdk import Context


class AlloyDBBaseOperator(GoogleCloudBaseOperator):
    """
    Base class for all AlloyDB operators.

    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "project_id",
        "location",
        "gcp_conn_id",
    )

    def __init__(
        self,
        project_id: str,
        location: str,
        gcp_conn_id: str = "google_cloud_default",
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        impersonation_chain: str | Sequence[str] | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata

    @cached_property
    def hook(self) -> AlloyDbHook:
        return AlloyDbHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )


class AlloyDBWriteBaseOperator(AlloyDBBaseOperator):
    """
    Base class for writing AlloyDB operators.

    These operators perform create, update or delete operations. with the objects (not inside of database).

    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    """

    template_fields: Sequence[str] = tuple(
        {"request_id", "validate_request"} | set(AlloyDBBaseOperator.template_fields)
    )

    def __init__(
        self,
        request_id: str | None = None,
        validate_request: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.request_id = request_id
        self.validate_request = validate_request

    def get_operation_result(self, operation: Operation) -> proto.Message | None:
        """
        Retrieve operation result as a proto.Message.

        If the `validate_request` parameter is set, then no operation is performed and thus nothing to wait.
        """
        if self.validate_request:
            # Validation requests are only validated and aren't executed, thus no operation result is expected
            return None
        return self.hook.wait_for_operation(timeout=self.timeout, operation=operation)


class AlloyDBCreateClusterOperator(AlloyDBWriteBaseOperator):
    """
    Create an Alloy DB cluster.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBCreateClusterOperator`

    :param cluster_id: Required. ID of the cluster to create.
    :param cluster_configuration: Required. Cluster to create. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Cluster
    :param is_secondary: Required. Specifies if the Cluster to be created is Primary or Secondary.
        Please note, if set True, then specify the `secondary_config` field in the cluster so the created
        secondary cluster was pointing to the primary cluster.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"cluster_id", "cluster_configuration", "is_secondary"}
        | set(AlloyDBWriteBaseOperator.template_fields)
    )
    operator_extra_links = (AlloyDBClusterLink(),)

    def __init__(
        self,
        cluster_id: str,
        cluster_configuration: alloydb_v1.Cluster | dict,
        is_secondary: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.cluster_id = cluster_id
        self.cluster_configuration = cluster_configuration
        self.is_secondary = is_secondary

    def _get_cluster(self) -> proto.Message | None:
        self.log.info("Checking if the cluster %s exists already...", self.cluster_id)
        try:
            cluster = self.hook.get_cluster(
                cluster_id=self.cluster_id,
                location=self.location,
                project_id=self.project_id,
            )
        except NotFound:
            self.log.info("The cluster %s does not exist yet.", self.cluster_id)
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            self.log.info("AlloyDB cluster %s already exists.", self.cluster_id)
            result = alloydb_v1.Cluster.to_dict(cluster)
            return result
        return None

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "location_id": self.location,
            "cluster_id": self.cluster_id,
            "project_id": self.project_id,
        }

    def execute(self, context: Context) -> dict | None:
        AlloyDBClusterLink.persist(context=context)
        if cluster := self._get_cluster():
            return cluster

        if self.validate_request:
            self.log.info("Validating a Create AlloyDB cluster request.")
        else:
            self.log.info("Creating an AlloyDB cluster.")

        try:
            create_method = (
                self.hook.create_secondary_cluster if self.is_secondary else self.hook.create_cluster
            )
            operation = create_method(
                cluster_id=self.cluster_id,
                cluster=self.cluster_configuration,
                location=self.location,
                project_id=self.project_id,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex)
        else:
            operation_result = self.get_operation_result(operation)
            result = alloydb_v1.Cluster.to_dict(operation_result) if operation_result else None

        return result


class AlloyDBUpdateClusterOperator(AlloyDBWriteBaseOperator):
    """
    Update an Alloy DB cluster.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBUpdateClusterOperator`

    :param cluster_id: Required. ID of the cluster to update.
    :param cluster_configuration: Required. Cluster to update. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Cluster
    :param update_mask: Optional. Field mask is used to specify the fields to be overwritten in the
            Cluster resource by the update.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param allow_missing: Optional. If set to true, update succeeds even if cluster is not found.
        In that case, a new cluster is created and update_mask is ignored.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"cluster_id", "cluster_configuration", "allow_missing"}
        | set(AlloyDBWriteBaseOperator.template_fields)
    )
    operator_extra_links = (AlloyDBClusterLink(),)

    def __init__(
        self,
        cluster_id: str,
        cluster_configuration: alloydb_v1.Cluster | dict,
        update_mask: FieldMask | dict | None = None,
        allow_missing: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.cluster_id = cluster_id
        self.cluster_configuration = cluster_configuration
        self.update_mask = update_mask
        self.allow_missing = allow_missing

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "location_id": self.location,
            "cluster_id": self.cluster_id,
            "project_id": self.project_id,
        }

    def execute(self, context: Context) -> dict | None:
        AlloyDBClusterLink.persist(context=context)
        if self.validate_request:
            self.log.info("Validating an Update AlloyDB cluster request.")
        else:
            self.log.info("Updating an AlloyDB cluster.")

        try:
            operation = self.hook.update_cluster(
                cluster_id=self.cluster_id,
                project_id=self.project_id,
                location=self.location,
                cluster=self.cluster_configuration,
                update_mask=self.update_mask,
                allow_missing=self.allow_missing,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            operation_result = self.get_operation_result(operation)
            result = alloydb_v1.Cluster.to_dict(operation_result) if operation_result else None

        if not self.validate_request:
            self.log.info("AlloyDB cluster %s was successfully updated.", self.cluster_id)
        return result


class AlloyDBDeleteClusterOperator(AlloyDBWriteBaseOperator):
    """
    Delete an Alloy DB cluster.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBDeleteClusterOperator`

    :param cluster_id: Required. ID of the cluster to delete.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param etag: Optional. The current etag of the Cluster. If an etag is provided and does not match the
            current etag of the Cluster, deletion will be blocked and an ABORTED error will be returned.
    :param force: Optional. Whether to cascade delete child instances for given cluster.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"cluster_id", "etag", "force"} | set(AlloyDBWriteBaseOperator.template_fields)
    )

    def __init__(
        self,
        cluster_id: str,
        etag: str | None = None,
        force: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.cluster_id = cluster_id
        self.etag = etag
        self.force = force

    def execute(self, context: Context) -> None:
        if self.validate_request:
            self.log.info("Validating a Delete AlloyDB cluster request.")
        else:
            self.log.info("Deleting an AlloyDB cluster.")

        try:
            operation = self.hook.delete_cluster(
                cluster_id=self.cluster_id,
                project_id=self.project_id,
                location=self.location,
                etag=self.etag,
                force=self.force,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            self.get_operation_result(operation)

        if not self.validate_request:
            self.log.info("AlloyDB cluster %s was successfully removed.", self.cluster_id)


class AlloyDBCreateInstanceOperator(AlloyDBWriteBaseOperator):
    """
    Create an Instance in an Alloy DB cluster.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBCreateInstanceOperator`

    :param cluster_id: Required. ID of the cluster for creating an instance in.
    :param instance_id: Required. ID of the instance to create.
    :param instance_configuration: Required. Instance to create. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Instance
    :param is_secondary: Required. Specifies if the Instance to be created is Primary or Secondary.
        Please note, if set True, then specify the `instance_type` field in the instance.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"cluster_id", "instance_id", "is_secondary", "instance_configuration"}
        | set(AlloyDBWriteBaseOperator.template_fields)
    )
    operator_extra_links = (AlloyDBClusterLink(),)

    def __init__(
        self,
        cluster_id: str,
        instance_id: str,
        instance_configuration: alloydb_v1.Instance | dict,
        is_secondary: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.cluster_id = cluster_id
        self.instance_id = instance_id
        self.instance_configuration = instance_configuration
        self.is_secondary = is_secondary

    def _get_instance(self) -> proto.Message | None:
        self.log.info("Checking if the instance %s exists already...", self.instance_id)
        try:
            instance = self.hook.get_instance(
                cluster_id=self.cluster_id,
                instance_id=self.instance_id,
                location=self.location,
                project_id=self.project_id,
            )
        except NotFound:
            self.log.info("The instance %s does not exist yet.", self.instance_id)
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            self.log.info(
                "AlloyDB instance %s already exists in the cluster %s.",
                self.cluster_id,
                self.instance_id,
            )
            result = alloydb_v1.Instance.to_dict(instance)
            return result
        return None

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "location_id": self.location,
            "cluster_id": self.cluster_id,
            "project_id": self.project_id,
        }

    def execute(self, context: Context) -> dict | None:
        AlloyDBClusterLink.persist(context=context)
        if instance := self._get_instance():
            return instance

        if self.validate_request:
            self.log.info("Validating a Create AlloyDB instance request.")
        else:
            self.log.info("Creating an AlloyDB instance.")

        try:
            create_method = (
                self.hook.create_secondary_instance if self.is_secondary else self.hook.create_instance
            )
            operation = create_method(
                cluster_id=self.cluster_id,
                instance_id=self.instance_id,
                instance=self.instance_configuration,
                location=self.location,
                project_id=self.project_id,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex)
        else:
            operation_result = self.get_operation_result(operation)
            result = alloydb_v1.Instance.to_dict(operation_result) if operation_result else None

        return result


class AlloyDBUpdateInstanceOperator(AlloyDBWriteBaseOperator):
    """
    Update an Alloy DB instance.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBUpdateInstanceOperator`

    :param cluster_id: Required. ID of the cluster.
    :param instance_id: Required. ID of the instance to update.
    :param instance_configuration: Required. Instance to update. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Instance
    :param update_mask: Optional. Field mask is used to specify the fields to be overwritten in the
            Instance resource by the update.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param allow_missing: Optional. If set to true, update succeeds even if instance is not found.
        In that case, a new instance is created and update_mask is ignored.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"cluster_id", "instance_id", "instance_configuration", "update_mask", "allow_missing"}
        | set(AlloyDBWriteBaseOperator.template_fields)
    )
    operator_extra_links = (AlloyDBClusterLink(),)

    def __init__(
        self,
        cluster_id: str,
        instance_id: str,
        instance_configuration: alloydb_v1.Instance | dict,
        update_mask: FieldMask | dict | None = None,
        allow_missing: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.cluster_id = cluster_id
        self.instance_id = instance_id
        self.instance_configuration = instance_configuration
        self.update_mask = update_mask
        self.allow_missing = allow_missing

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "location_id": self.location,
            "cluster_id": self.cluster_id,
            "project_id": self.project_id,
        }

    def execute(self, context: Context) -> dict | None:
        AlloyDBClusterLink.persist(context=context)
        if self.validate_request:
            self.log.info("Validating an Update AlloyDB instance request.")
        else:
            self.log.info("Updating an AlloyDB instance.")

        try:
            operation = self.hook.update_instance(
                cluster_id=self.cluster_id,
                instance_id=self.instance_id,
                project_id=self.project_id,
                location=self.location,
                instance=self.instance_configuration,
                update_mask=self.update_mask,
                allow_missing=self.allow_missing,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            operation_result = self.get_operation_result(operation)
            result = alloydb_v1.Instance.to_dict(operation_result) if operation_result else None

        if not self.validate_request:
            self.log.info("AlloyDB instance %s was successfully updated.", self.cluster_id)
        return result


class AlloyDBDeleteInstanceOperator(AlloyDBWriteBaseOperator):
    """
    Delete an Alloy DB instance.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBDeleteInstanceOperator`

    :param instance_id: Required. ID of the instance to delete.
    :param cluster_id: Required. ID of the cluster.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param etag: Optional. The current etag of the Instance. If an etag is provided and does not match the
            current etag of the Instance, deletion will be blocked and an ABORTED error will be returned.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"instance_id", "cluster_id", "etag"} | set(AlloyDBWriteBaseOperator.template_fields)
    )

    def __init__(
        self,
        instance_id: str,
        cluster_id: str,
        etag: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.instance_id = instance_id
        self.cluster_id = cluster_id
        self.etag = etag

    def execute(self, context: Context) -> None:
        if self.validate_request:
            self.log.info("Validating a Delete AlloyDB instance request.")
        else:
            self.log.info("Deleting an AlloyDB instance.")

        try:
            operation = self.hook.delete_instance(
                instance_id=self.instance_id,
                cluster_id=self.cluster_id,
                project_id=self.project_id,
                location=self.location,
                etag=self.etag,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            self.get_operation_result(operation)

        if not self.validate_request:
            self.log.info("AlloyDB instance %s was successfully removed.", self.instance_id)


class AlloyDBCreateUserOperator(AlloyDBWriteBaseOperator):
    """
    Create a User in an Alloy DB cluster.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBCreateUserOperator`

    :param user_id: Required. ID of the user to create.
    :param user_configuration: Required. The user to create. For more details please see API documentation:
        https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.User
    :param cluster_id: Required. ID of the cluster for creating a user in.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"user_id", "user_configuration", "cluster_id"} | set(AlloyDBWriteBaseOperator.template_fields)
    )
    operator_extra_links = (AlloyDBUsersLink(),)

    def __init__(
        self,
        user_id: str,
        user_configuration: alloydb_v1.User | dict,
        cluster_id: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.user_id = user_id
        self.user_configuration = user_configuration
        self.cluster_id = cluster_id

    def _get_user(self) -> proto.Message | None:
        self.log.info("Checking if the user %s exists already...", self.user_id)
        try:
            user = self.hook.get_user(
                user_id=self.user_id,
                cluster_id=self.cluster_id,
                location=self.location,
                project_id=self.project_id,
            )
        except NotFound:
            self.log.info("The user %s does not exist yet.", self.user_id)
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            self.log.info(
                "AlloyDB user %s already exists in the cluster %s.",
                self.user_id,
                self.cluster_id,
            )
            result = alloydb_v1.User.to_dict(user)
            return result
        return None

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "location_id": self.location,
            "cluster_id": self.cluster_id,
            "project_id": self.project_id,
        }

    def execute(self, context: Context) -> dict | None:
        AlloyDBUsersLink.persist(context=context)
        if (_user := self._get_user()) is not None:
            return _user

        if self.validate_request:
            self.log.info("Validating a Create AlloyDB user request.")
        else:
            self.log.info("Creating an AlloyDB user.")

        try:
            user = self.hook.create_user(
                user_id=self.user_id,
                cluster_id=self.cluster_id,
                user=self.user_configuration,
                location=self.location,
                project_id=self.project_id,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex)
        else:
            result = alloydb_v1.User.to_dict(user) if not self.validate_request else None

        if not self.validate_request:
            self.log.info("AlloyDB user %s was successfully created.", self.user_id)
        return result


class AlloyDBUpdateUserOperator(AlloyDBWriteBaseOperator):
    """
    Update an Alloy DB user.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBUpdateUserOperator`

    :param user_id: Required. The ID of the user to update.
    :param cluster_id: Required. ID of the cluster.
    :param user_configuration: Required. User to update. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.User
    :param update_mask: Optional. Field mask is used to specify the fields to be overwritten in the
            User resource by the update.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param allow_missing: Optional. If set to true, update succeeds even if instance is not found.
        In that case, a new user is created and update_mask is ignored.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"cluster_id", "user_id", "user_configuration", "update_mask", "allow_missing"}
        | set(AlloyDBWriteBaseOperator.template_fields)
    )
    operator_extra_links = (AlloyDBUsersLink(),)

    def __init__(
        self,
        cluster_id: str,
        user_id: str,
        user_configuration: alloydb_v1.User | dict,
        update_mask: FieldMask | dict | None = None,
        allow_missing: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.cluster_id = cluster_id
        self.user_id = user_id
        self.user_configuration = user_configuration
        self.update_mask = update_mask
        self.allow_missing = allow_missing

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "location_id": self.location,
            "cluster_id": self.cluster_id,
            "project_id": self.project_id,
        }

    def execute(self, context: Context) -> dict | None:
        AlloyDBUsersLink.persist(context=context)
        if self.validate_request:
            self.log.info("Validating an Update AlloyDB user request.")
        else:
            self.log.info("Updating an AlloyDB user.")

        try:
            user = self.hook.update_user(
                cluster_id=self.cluster_id,
                user_id=self.user_id,
                project_id=self.project_id,
                location=self.location,
                user=self.user_configuration,
                update_mask=self.update_mask,
                allow_missing=self.allow_missing,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            result = alloydb_v1.User.to_dict(user) if not self.validate_request else None

        if not self.validate_request:
            self.log.info("AlloyDB user %s was successfully updated.", self.user_id)
        return result


class AlloyDBDeleteUserOperator(AlloyDBWriteBaseOperator):
    """
    Delete an Alloy DB user.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBDeleteUserOperator`

    :param user_id: Required. ID of the user to delete.
    :param cluster_id: Required. ID of the cluster.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"user_id", "cluster_id"} | set(AlloyDBWriteBaseOperator.template_fields)
    )

    def __init__(
        self,
        user_id: str,
        cluster_id: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.user_id = user_id
        self.cluster_id = cluster_id

    def execute(self, context: Context) -> None:
        if self.validate_request:
            self.log.info("Validating a Delete AlloyDB user request.")
        else:
            self.log.info("Deleting an AlloyDB user.")

        try:
            self.hook.delete_user(
                user_id=self.user_id,
                cluster_id=self.cluster_id,
                project_id=self.project_id,
                location=self.location,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex) from ex

        if not self.validate_request:
            self.log.info("AlloyDB user %s was successfully removed.", self.user_id)


class AlloyDBCreateBackupOperator(AlloyDBWriteBaseOperator):
    """
    Create a Backup in an Alloy DB cluster.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBCreateBackupOperator`

    :param backup_id: Required. ID of the backup to create.
    :param backup_configuration: Required. Backup to create. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Backup
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the backups should be saved.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"backup_id", "backup_configuration"} | set(AlloyDBWriteBaseOperator.template_fields)
    )
    operator_extra_links = (AlloyDBBackupsLink(),)

    def __init__(
        self,
        backup_id: str,
        backup_configuration: alloydb_v1.Backup | dict,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.backup_id = backup_id
        self.backup_configuration = backup_configuration

    def _get_backup(self) -> proto.Message | None:
        self.log.info("Checking if the backup %s exists already...", self.backup_id)
        try:
            backup = self.hook.get_backup(
                backup_id=self.backup_id,
                location=self.location,
                project_id=self.project_id,
            )
        except NotFound:
            self.log.info("The backup %s does not exist yet.", self.backup_id)
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            self.log.info("AlloyDB backup %s already exists.", self.backup_id)
            result = alloydb_v1.Backup.to_dict(backup)
            return result
        return None

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "project_id": self.project_id,
        }

    def execute(self, context: Context) -> dict | None:
        AlloyDBBackupsLink.persist(context=context)
        if backup := self._get_backup():
            return backup

        if self.validate_request:
            self.log.info("Validating a Create AlloyDB backup request.")
        else:
            self.log.info("Creating an AlloyDB backup.")

        try:
            operation = self.hook.create_backup(
                backup_id=self.backup_id,
                backup=self.backup_configuration,
                location=self.location,
                project_id=self.project_id,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex)
        else:
            operation_result = self.get_operation_result(operation)
            result = alloydb_v1.Backup.to_dict(operation_result) if operation_result else None

        return result


class AlloyDBUpdateBackupOperator(AlloyDBWriteBaseOperator):
    """
    Update an Alloy DB backup.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBUpdateBackupOperator`

    :param backup_id: Required. ID of the backup to update.
    :param backup_configuration: Required. Backup to update. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Backup
    :param update_mask: Optional. Field mask is used to specify the fields to be overwritten in the
            Backup resource by the update.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param allow_missing: Optional. If set to true, update succeeds even if backup is not found.
        In that case, a new backup is created and update_mask is ignored.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"backup_id", "backup_configuration", "update_mask", "allow_missing"}
        | set(AlloyDBWriteBaseOperator.template_fields)
    )
    operator_extra_links = (AlloyDBBackupsLink(),)

    def __init__(
        self,
        backup_id: str,
        backup_configuration: alloydb_v1.Backup | dict,
        update_mask: FieldMask | dict | None = None,
        allow_missing: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.backup_id = backup_id
        self.backup_configuration = backup_configuration
        self.update_mask = update_mask
        self.allow_missing = allow_missing

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "project_id": self.project_id,
        }

    def execute(self, context: Context) -> dict | None:
        AlloyDBBackupsLink.persist(context=context)
        if self.validate_request:
            self.log.info("Validating an Update AlloyDB backup request.")
        else:
            self.log.info("Updating an AlloyDB backup.")

        try:
            operation = self.hook.update_backup(
                backup_id=self.backup_id,
                project_id=self.project_id,
                location=self.location,
                backup=self.backup_configuration,
                update_mask=self.update_mask,
                allow_missing=self.allow_missing,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            operation_result = self.get_operation_result(operation)
            result = alloydb_v1.Backup.to_dict(operation_result) if operation_result else None

        if not self.validate_request:
            self.log.info("AlloyDB backup %s was successfully updated.", self.backup_id)
        return result


class AlloyDBDeleteBackupOperator(AlloyDBWriteBaseOperator):
    """
    Delete an Alloy DB backup.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AlloyDBDeleteBackupOperator`

    :param backup_id: Required. ID of the backup to delete.
    :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
        so that if you must retry your request, the server ignores the request if it has already been
        completed. The server guarantees that for at least 60 minutes since the first request.
        For example, consider a situation where you make an initial request and the request times out.
        If you make the request again with the same request ID, the server can check if the original operation
        with the same request ID was received, and if so, ignores the second request.
        This prevents clients from accidentally creating duplicate commitments.
        The request ID must be a valid UUID with the exception that zero UUID is not supported
        (00000000-0000-0000-0000-000000000000).
    :param validate_request: Optional. If set, performs request validation, but does not actually
        execute the request.
    :param project_id: Required. The ID of the Google Cloud project where the service is used.
    :param location: Required. The ID of the Google Cloud region where the service is used.
    :param gcp_conn_id: Optional. The connection ID to use to connect to Google Cloud.
    :param retry: Optional. A retry object used to retry requests. If `None` is specified, requests will not
        be retried.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Note that if `retry` is specified, the timeout applies to each individual attempt.
    :param metadata: Optional. Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple({"backup_id"} | set(AlloyDBWriteBaseOperator.template_fields))

    def __init__(
        self,
        backup_id: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.backup_id = backup_id

    def execute(self, context: Context) -> None:
        if self.validate_request:
            self.log.info("Validating a Delete AlloyDB backup request.")
        else:
            self.log.info("Deleting an AlloyDB backup.")

        try:
            operation = self.hook.delete_backup(
                backup_id=self.backup_id,
                project_id=self.project_id,
                location=self.location,
                request_id=self.request_id,
                validate_only=self.validate_request,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
        except Exception as ex:
            raise AirflowException(ex) from ex
        else:
            self.get_operation_result(operation)

        if not self.validate_request:
            self.log.info("AlloyDB backup %s was successfully removed.", self.backup_id)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import PropertyMock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.functions import CloudFunctionsHook

from unit.google.cloud.utils.base_gcp_mock import (
    GCP_PROJECT_ID_HOOK_UNIT_TEST,
    get_open_mock,
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

GCF_LOCATION = "location"
GCF_FUNCTION = "function"


class TestFunctionHookNoDefaultProjectId:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_no_default_project_id,
        ):
            self.gcf_function_hook_no_project_id = CloudFunctionsHook(gcp_conn_id="test", api_version="v1")

    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook._authorize")
    @mock.patch("airflow.providers.google.cloud.hooks.functions.build")
    def test_gcf_client_creation(self, mock_build, mock_authorize):
        result = self.gcf_function_hook_no_project_id.get_conn()
        mock_build.assert_called_once_with(
            "cloudfunctions", "v1", http=mock_authorize.return_value, cache_discovery=False
        )
        assert mock_build.return_value == result
        assert self.gcf_function_hook_no_project_id._conn == result

    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook._wait_for_operation_to_complete"
    )
    def test_create_new_function_overridden_project_id(self, wait_for_operation_to_complete, get_conn):
        create_method = (
            get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.create
        )
        execute_method = create_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gcf_function_hook_no_project_id.create_new_function(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST, location=GCF_LOCATION, body={}
        )
        assert res is None
        create_method.assert_called_once_with(body={}, location="projects/example-project/locations/location")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(operation_name="operation_id")

    @mock.patch("requests.put")
    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    def test_upload_function_zip_overridden_project_id(self, get_conn, requests_put):
        mck, open_module = get_open_mock()
        with mock.patch(f"{open_module}.open", mck):
            generate_upload_url_method = get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.generateUploadUrl

            execute_method = generate_upload_url_method.return_value.execute
            execute_method.return_value = {"uploadUrl": "http://uploadHere"}
            requests_put.return_value = None
            res = self.gcf_function_hook_no_project_id.upload_function_zip(
                project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST, location=GCF_LOCATION, zip_path="/tmp/path.zip"
            )
            assert res == "http://uploadHere"
            generate_upload_url_method.assert_called_once_with(
                parent="projects/example-project/locations/location"
            )
            execute_method.assert_called_once_with(num_retries=5)
            requests_put.assert_called_once_with(
                data=mock.ANY,
                headers={"Content-type": "application/zip", "x-goog-content-length-range": "0,104857600"},
                url="http://uploadHere",
            )


class TestFunctionHookDefaultProjectId:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.gcf_function_hook = CloudFunctionsHook(gcp_conn_id="test", api_version="v1")

    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook._authorize")
    @mock.patch("airflow.providers.google.cloud.hooks.functions.build")
    def test_gcf_client_creation(self, mock_build, mock_authorize):
        result = self.gcf_function_hook.get_conn()
        mock_build.assert_called_once_with(
            "cloudfunctions", "v1", http=mock_authorize.return_value, cache_discovery=False
        )
        assert mock_build.return_value == result
        assert self.gcf_function_hook._conn == result

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook._wait_for_operation_to_complete"
    )
    def test_create_new_function(self, wait_for_operation_to_complete, get_conn, mock_project_id):
        create_method = (
            get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.create
        )
        execute_method = create_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gcf_function_hook.create_new_function(
            location=GCF_LOCATION,
            body={},
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        assert res is None
        create_method.assert_called_once_with(body={}, location="projects/example-project/locations/location")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(operation_name="operation_id")

    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook._wait_for_operation_to_complete"
    )
    def test_create_new_function_override_project_id(self, wait_for_operation_to_complete, get_conn):
        create_method = (
            get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.create
        )
        execute_method = create_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gcf_function_hook.create_new_function(
            project_id="new-project", location=GCF_LOCATION, body={}
        )
        assert res is None
        create_method.assert_called_once_with(body={}, location="projects/new-project/locations/location")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(operation_name="operation_id")

    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    def test_get_function(self, get_conn):
        get_method = (
            get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.get
        )
        execute_method = get_method.return_value.execute
        execute_method.return_value = {"name": "function"}
        res = self.gcf_function_hook.get_function(name=GCF_FUNCTION)
        assert res is not None
        assert res["name"] == "function"
        get_method.assert_called_once_with(name="function")
        execute_method.assert_called_once_with(num_retries=5)

    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook._wait_for_operation_to_complete"
    )
    def test_delete_function(self, wait_for_operation_to_complete, get_conn):
        delete_method = (
            get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.delete
        )
        execute_method = delete_method.return_value.execute
        wait_for_operation_to_complete.return_value = None
        execute_method.return_value = {"name": "operation_id"}
        res = self.gcf_function_hook.delete_function(name=GCF_FUNCTION)
        assert res is None
        delete_method.assert_called_once_with(name="function")
        execute_method.assert_called_once_with(num_retries=5)

    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook._wait_for_operation_to_complete"
    )
    def test_update_function(self, wait_for_operation_to_complete, get_conn):
        patch_method = (
            get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.patch
        )
        execute_method = patch_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gcf_function_hook.update_function(update_mask=["a", "b", "c"], name=GCF_FUNCTION, body={})
        assert res is None
        patch_method.assert_called_once_with(body={}, name="function", updateMask="a,b,c")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(operation_name="operation_id")

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("requests.put")
    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    def test_upload_function_zip(self, get_conn, requests_put, mock_project_id):
        mck, open_module = get_open_mock()
        with mock.patch(f"{open_module}.open", mck):
            generate_upload_url_method = get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.generateUploadUrl

            execute_method = generate_upload_url_method.return_value.execute
            execute_method.return_value = {"uploadUrl": "http://uploadHere"}
            requests_put.return_value = None
            res = self.gcf_function_hook.upload_function_zip(
                location=GCF_LOCATION,
                zip_path="/tmp/path.zip",
                project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            )
            assert res == "http://uploadHere"
            generate_upload_url_method.assert_called_once_with(
                parent="projects/example-project/locations/location"
            )
            execute_method.assert_called_once_with(num_retries=5)
            requests_put.assert_called_once_with(
                data=mock.ANY,
                headers={"Content-type": "application/zip", "x-goog-content-length-range": "0,104857600"},
                url="http://uploadHere",
            )

    @mock.patch("requests.put")
    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    def test_upload_function_zip_overridden_project_id(self, get_conn, requests_put):
        mck, open_module = get_open_mock()
        with mock.patch(f"{open_module}.open", mck):
            generate_upload_url_method = get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.generateUploadUrl

            execute_method = generate_upload_url_method.return_value.execute
            execute_method.return_value = {"uploadUrl": "http://uploadHere"}
            requests_put.return_value = None
            res = self.gcf_function_hook.upload_function_zip(
                project_id="new-project", location=GCF_LOCATION, zip_path="/tmp/path.zip"
            )
            assert res == "http://uploadHere"
            generate_upload_url_method.assert_called_once_with(
                parent="projects/new-project/locations/location"
            )
            execute_method.assert_called_once_with(num_retries=5)
            requests_put.assert_called_once_with(
                data=mock.ANY,
                headers={"Content-type": "application/zip", "x-goog-content-length-range": "0,104857600"},
                url="http://uploadHere",
            )

    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    def test_call_function(self, mock_get_conn):
        payload = {"executionId": "wh41ppcyoa6l", "result": "Hello World!"}

        call = mock_get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.call

        call.return_value.execute.return_value = payload

        function_id = "function1234"
        input_data = {"key": "value"}
        name = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/locations/{GCF_LOCATION}/functions/{function_id}"

        result = self.gcf_function_hook.call_function(
            function_id=function_id,
            location=GCF_LOCATION,
            input_data=input_data,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )

        call.assert_called_once_with(body=input_data, name=name)
        assert result == payload

    @mock.patch("airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook.get_conn")
    def test_call_function_error(self, mock_get_conn):
        payload = {"error": "Something very bad"}

        call = mock_get_conn.return_value.projects.return_value.locations.return_value.functions.return_value.call

        call.return_value.execute.return_value = payload

        function_id = "function1234"
        input_data = {"key": "value"}
        with pytest.raises(AirflowException):
            self.gcf_function_hook.call_function(
                function_id=function_id,
                location=GCF_LOCATION,
                input_data=input_data,
                project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from unittest import mock

import pytest
from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.monitoring_v3 import AlertPolicy, NotificationChannel
from google.protobuf.field_mask_pb2 import FieldMask

from airflow.providers.google.cloud.hooks import stackdriver

PROJECT_ID = "sd-project"
CREDENTIALS = "sd-credentials"
TEST_FILTER = "filter"
TEST_ALERT_POLICY_1 = {
    "combiner": "OR",
    "name": "projects/sd-project/alertPolicies/12345",
    "enabled": True,
    "display_name": "test display",
    "conditions": [
        {
            "condition_threshold": {
                "comparison": "COMPARISON_GT",
                "aggregations": [{"alignment_period": {"seconds": 60}, "per_series_aligner": "ALIGN_RATE"}],
            },
            "display_name": "Condition display",
            "name": "projects/sd-project/alertPolicies/123/conditions/456",
        }
    ],
}

TEST_ALERT_POLICY_2 = {
    "combiner": "OR",
    "name": "projects/sd-project/alertPolicies/6789",
    "enabled": False,
    "display_name": "test display",
    "conditions": [
        {
            "condition_threshold": {
                "comparison": "COMPARISON_GT",
                "aggregations": [{"alignment_period": {"seconds": 60}, "per_series_aligner": "ALIGN_RATE"}],
            },
            "display_name": "Condition display",
            "name": "projects/sd-project/alertPolicies/456/conditions/789",
        }
    ],
}

TEST_NOTIFICATION_CHANNEL_1 = {
    "display_name": "sd",
    "enabled": True,
    "labels": {"auth_token": "top-secret", "channel_name": "#channel"},
    "name": "projects/sd-project/notificationChannels/12345",
    "type_": "slack",
}

TEST_NOTIFICATION_CHANNEL_2 = {
    "display_name": "sd",
    "enabled": False,
    "labels": {"auth_token": "top-secret", "channel_name": "#channel"},
    "name": "projects/sd-project/notificationChannels/6789",
    "type_": "slack",
}


@pytest.mark.db_test
class TestStackdriverHookMethods:
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_policy_client")
    def test_stackdriver_list_alert_policies(self, mock_policy_client, mock_get_creds_and_project_id):
        method = mock_policy_client.return_value.list_alert_policies
        hook = stackdriver.StackdriverHook()
        hook.list_alert_policies(
            filter_=TEST_FILTER,
            project_id=PROJECT_ID,
        )
        method.assert_called_once_with(
            request=dict(name=f"projects/{PROJECT_ID}", filter=TEST_FILTER, order_by=None, page_size=None),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_policy_client")
    def test_stackdriver_enable_alert_policy(self, mock_policy_client, mock_get_creds_and_project_id):
        hook = stackdriver.StackdriverHook()

        alert_policy_enabled = AlertPolicy(**TEST_ALERT_POLICY_1)
        alert_policy_disabled = AlertPolicy(**TEST_ALERT_POLICY_2)

        alert_policies = [alert_policy_enabled, alert_policy_disabled]

        mock_policy_client.return_value.list_alert_policies.return_value = alert_policies
        hook.enable_alert_policies(
            filter_=TEST_FILTER,
            project_id=PROJECT_ID,
        )
        mock_policy_client.return_value.list_alert_policies.assert_called_once_with(
            request=dict(name=f"projects/{PROJECT_ID}", filter=TEST_FILTER, order_by=None, page_size=None),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        mask = FieldMask(paths=["enabled"])
        alert_policy_disabled.enabled = True
        mock_policy_client.return_value.update_alert_policy.assert_called_once_with(
            request=dict(alert_policy=alert_policy_disabled, update_mask=mask),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_policy_client")
    def test_stackdriver_disable_alert_policy(self, mock_policy_client, mock_get_creds_and_project_id):
        hook = stackdriver.StackdriverHook()
        alert_policy_enabled = AlertPolicy(**TEST_ALERT_POLICY_1)
        alert_policy_disabled = AlertPolicy(**TEST_ALERT_POLICY_2)

        mock_policy_client.return_value.list_alert_policies.return_value = [
            alert_policy_enabled,
            alert_policy_disabled,
        ]
        hook.disable_alert_policies(
            filter_=TEST_FILTER,
            project_id=PROJECT_ID,
        )
        mock_policy_client.return_value.list_alert_policies.assert_called_once_with(
            request=dict(name=f"projects/{PROJECT_ID}", filter=TEST_FILTER, order_by=None, page_size=None),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        mask = FieldMask(paths=["enabled"])
        alert_policy_enabled.enabled = False
        mock_policy_client.return_value.update_alert_policy.assert_called_once_with(
            request=dict(alert_policy=alert_policy_enabled, update_mask=mask),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_policy_client")
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_channel_client")
    def test_stackdriver_upsert_alert_policy(
        self, mock_channel_client, mock_policy_client, mock_get_creds_and_project_id
    ):
        hook = stackdriver.StackdriverHook()
        existing_alert_policy = AlertPolicy(**TEST_ALERT_POLICY_1)
        alert_policy_to_create = AlertPolicy(**TEST_ALERT_POLICY_2)

        mock_policy_client.return_value.list_alert_policies.return_value = [existing_alert_policy]
        mock_channel_client.return_value.list_notification_channels.return_value = []

        hook.upsert_alert(
            alerts=json.dumps({"policies": [TEST_ALERT_POLICY_1, TEST_ALERT_POLICY_2], "channels": []}),
            project_id=PROJECT_ID,
        )
        mock_channel_client.return_value.list_notification_channels.assert_called_once_with(
            request=dict(
                name=f"projects/{PROJECT_ID}",
                filter=None,
                order_by=None,
                page_size=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        mock_policy_client.return_value.list_alert_policies.assert_called_once_with(
            request=dict(name=f"projects/{PROJECT_ID}", filter=None, order_by=None, page_size=None),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        alert_policy_to_create.name = None
        alert_policy_to_create.creation_record = None
        alert_policy_to_create.mutation_record = None

        alert_policy_to_create.conditions[0].name = None
        mock_policy_client.return_value.create_alert_policy.assert_called_once_with(
            request=dict(
                name=f"projects/{PROJECT_ID}",
                alert_policy=alert_policy_to_create,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        existing_alert_policy.creation_record = None
        existing_alert_policy.mutation_record = None
        mock_policy_client.return_value.update_alert_policy.assert_called_once_with(
            request=dict(alert_policy=existing_alert_policy), retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_policy_client")
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_channel_client")
    def test_stackdriver_upsert_alert_policy_without_channel(
        self, mock_channel_client, mock_policy_client, mock_get_creds_and_project_id
    ):
        hook = stackdriver.StackdriverHook()
        existing_alert_policy = AlertPolicy(**TEST_ALERT_POLICY_1)

        mock_policy_client.return_value.list_alert_policies.return_value = [existing_alert_policy]
        mock_channel_client.return_value.list_notification_channels.return_value = []

        hook.upsert_alert(
            alerts=json.dumps({"policies": [TEST_ALERT_POLICY_1, TEST_ALERT_POLICY_2]}),
            project_id=PROJECT_ID,
        )
        mock_channel_client.return_value.list_notification_channels.assert_called_once_with(
            request=dict(name=f"projects/{PROJECT_ID}", filter=None, order_by=None, page_size=None),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_policy_client.return_value.list_alert_policies.assert_called_once_with(
            request=dict(name=f"projects/{PROJECT_ID}", filter=None, order_by=None, page_size=None),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

        existing_alert_policy.creation_record = None
        existing_alert_policy.mutation_record = None
        mock_policy_client.return_value.update_alert_policy.assert_called_once_with(
            request=dict(alert_policy=existing_alert_policy), retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_policy_client")
    def test_stackdriver_delete_alert_policy(self, mock_policy_client, mock_get_creds_and_project_id):
        hook = stackdriver.StackdriverHook()
        hook.delete_alert_policy(
            name="test-alert",
        )
        mock_policy_client.return_value.delete_alert_policy.assert_called_once_with(
            request=dict(name="test-alert"),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_channel_client")
    def test_stackdriver_list_notification_channel(self, mock_channel_client, mock_get_creds_and_project_id):
        hook = stackdriver.StackdriverHook()
        hook.list_notification_channels(
            filter_=TEST_FILTER,
            project_id=PROJECT_ID,
        )
        mock_channel_client.return_value.list_notification_channels.assert_called_once_with(
            request=dict(name=f"projects/{PROJECT_ID}", filter=TEST_FILTER, order_by=None, page_size=None),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_channel_client")
    def test_stackdriver_enable_notification_channel(
        self, mock_channel_client, mock_get_creds_and_project_id
    ):
        hook = stackdriver.StackdriverHook()
        notification_channel_enabled = NotificationChannel(**TEST_NOTIFICATION_CHANNEL_1)
        notification_channel_disabled = NotificationChannel(**TEST_NOTIFICATION_CHANNEL_2)

        mock_channel_client.return_value.list_notification_channels.return_value = [
            notification_channel_enabled,
            notification_channel_disabled,
        ]

        hook.enable_notification_channels(
            filter_=TEST_FILTER,
            project_id=PROJECT_ID,
        )

        notification_channel_disabled.enabled = True
        mask = FieldMask(paths=["enabled"])
        mock_channel_client.return_value.update_notification_channel.assert_called_once_with(
            request=dict(notification_channel=notification_channel_disabled, update_mask=mask),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_channel_client")
    def test_stackdriver_disable_notification_channel(
        self, mock_channel_client, mock_get_creds_and_project_id
    ):
        hook = stackdriver.StackdriverHook()
        notification_channel_enabled = NotificationChannel(**TEST_NOTIFICATION_CHANNEL_1)
        notification_channel_disabled = NotificationChannel(**TEST_NOTIFICATION_CHANNEL_2)
        mock_channel_client.return_value.list_notification_channels.return_value = [
            notification_channel_enabled,
            notification_channel_disabled,
        ]

        hook.disable_notification_channels(
            filter_=TEST_FILTER,
            project_id=PROJECT_ID,
        )

        notification_channel_enabled.enabled = False
        mask = FieldMask(paths=["enabled"])
        mock_channel_client.return_value.update_notification_channel.assert_called_once_with(
            request=dict(notification_channel=notification_channel_enabled, update_mask=mask),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_channel_client")
    def test_stackdriver_upsert_channel(self, mock_channel_client, mock_get_creds_and_project_id):
        hook = stackdriver.StackdriverHook()
        existing_notification_channel = NotificationChannel(**TEST_NOTIFICATION_CHANNEL_1)
        notification_channel_to_be_created = NotificationChannel(**TEST_NOTIFICATION_CHANNEL_2)

        mock_channel_client.return_value.list_notification_channels.return_value = [
            existing_notification_channel
        ]
        hook.upsert_channel(
            channels=json.dumps({"channels": [TEST_NOTIFICATION_CHANNEL_1, TEST_NOTIFICATION_CHANNEL_2]}),
            project_id=PROJECT_ID,
        )
        mock_channel_client.return_value.list_notification_channels.assert_called_once_with(
            request=dict(name=f"projects/{PROJECT_ID}", filter=None, order_by=None, page_size=None),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        mock_channel_client.return_value.update_notification_channel.assert_called_once_with(
            request=dict(notification_channel=existing_notification_channel),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        notification_channel_to_be_created.name = None
        mock_channel_client.return_value.create_notification_channel.assert_called_once_with(
            request=dict(
                name=f"projects/{PROJECT_ID}", notification_channel=notification_channel_to_be_created
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id",
        return_value=(CREDENTIALS, PROJECT_ID),
    )
    @mock.patch("airflow.providers.google.cloud.hooks.stackdriver.StackdriverHook._get_channel_client")
    def test_stackdriver_delete_notification_channel(
        self, mock_channel_client, mock_get_creds_and_project_id
    ):
        hook = stackdriver.StackdriverHook()
        hook.delete_notification_channel(
            name="test-channel",
        )
        mock_channel_client.return_value.delete_notification_channel.assert_called_once_with(
            request=dict(name="test-channel"), retry=DEFAULT, timeout=None, metadata=()
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any

from google.cloud.dataflow_v1beta3 import JobState
from google.cloud.dataflow_v1beta3.types import (
    AutoscalingEvent,
    Job,
    JobMessage,
    JobMetrics,
    JobType,
    MetricUpdate,
)

from airflow.providers.google.cloud.hooks.dataflow import AsyncDataflowHook, DataflowJobStatus
from airflow.triggers.base import BaseTrigger, TriggerEvent

if TYPE_CHECKING:
    from google.cloud.dataflow_v1beta3.services.messages_v1_beta3.pagers import ListJobMessagesAsyncPager


DEFAULT_DATAFLOW_LOCATION = "us-central1"


class TemplateJobStartTrigger(BaseTrigger):
    """
    Dataflow trigger to check if templated job has been finished.

    :param project_id: Required. the Google Cloud project ID in which the job was started.
    :param job_id: Required. ID of the job.
    :param location: Optional. the location where job is executed. If set to None then
        the value of DEFAULT_DATAFLOW_LOCATION will be used
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional. Service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param cancel_timeout: Optional. How long (in seconds) operator should wait for the pipeline to be
        successfully cancelled when task is being killed.
    """

    def __init__(
        self,
        job_id: str,
        project_id: str | None,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        poll_sleep: int = 10,
        impersonation_chain: str | Sequence[str] | None = None,
        cancel_timeout: int | None = 5 * 60,
    ):
        super().__init__()
        self.project_id = project_id
        self.job_id = job_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.poll_sleep = poll_sleep
        self.impersonation_chain = impersonation_chain
        self.cancel_timeout = cancel_timeout

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize class arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.dataflow.TemplateJobStartTrigger",
            {
                "project_id": self.project_id,
                "job_id": self.job_id,
                "location": self.location,
                "gcp_conn_id": self.gcp_conn_id,
                "poll_sleep": self.poll_sleep,
                "impersonation_chain": self.impersonation_chain,
                "cancel_timeout": self.cancel_timeout,
            },
        )

    async def run(self):
        """
        Fetch job status or yield certain Events.

        Main loop of the class in where it is fetching the job status and yields certain Event.

        If the job has status success then it yields TriggerEvent with success status, if job has
        status failed - with error status. In any other case Trigger will wait for specified
        amount of time stored in self.poll_sleep variable.
        """
        hook = self._get_async_hook()
        try:
            while True:
                status = await hook.get_job_status(
                    project_id=self.project_id,
                    job_id=self.job_id,
                    location=self.location,
                )
                if status == JobState.JOB_STATE_DONE:
                    yield TriggerEvent(
                        {
                            "job_id": self.job_id,
                            "status": "success",
                            "message": "Job completed",
                        }
                    )
                    return
                elif status == JobState.JOB_STATE_FAILED:
                    yield TriggerEvent(
                        {
                            "status": "error",
                            "message": f"Dataflow job with id {self.job_id} has failed its execution",
                        }
                    )
                    return
                elif status == JobState.JOB_STATE_STOPPED:
                    yield TriggerEvent(
                        {
                            "status": "stopped",
                            "message": f"Dataflow job with id {self.job_id} was stopped",
                        }
                    )
                    return
                else:
                    self.log.info("Job is still running...")
                    self.log.info("Current job status is: %s", status.name)
                    self.log.info("Sleeping for %s seconds.", self.poll_sleep)
                    await asyncio.sleep(self.poll_sleep)
        except Exception as e:
            self.log.exception("Exception occurred while checking for job completion.")
            yield TriggerEvent({"status": "error", "message": str(e)})

    def _get_async_hook(self) -> AsyncDataflowHook:
        return AsyncDataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            poll_sleep=self.poll_sleep,
            impersonation_chain=self.impersonation_chain,
            cancel_timeout=self.cancel_timeout,
        )


class DataflowJobStatusTrigger(BaseTrigger):
    """
    Trigger that monitors if a Dataflow job has reached any of the expected statuses.

    :param job_id: Required. ID of the job.
    :param expected_statuses: The expected state(s) of the operation.
        See: https://cloud.google.com/dataflow/docs/reference/rest/v1b3/projects.jobs#Job.JobState
    :param project_id: Required. The Google Cloud project ID in which the job was started.
    :param location: Optional. The location where the job is executed. If set to None then
        the value of DEFAULT_DATAFLOW_LOCATION will be used.
    :param gcp_conn_id: The connection ID to use for connecting to Google Cloud.
    :param poll_sleep: Time (seconds) to wait between two consecutive calls to check the job.
    :param impersonation_chain: Optional. Service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    def __init__(
        self,
        job_id: str,
        expected_statuses: set[str],
        project_id: str | None,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        poll_sleep: int = 10,
        impersonation_chain: str | Sequence[str] | None = None,
    ):
        super().__init__()
        self.job_id = job_id
        self.expected_statuses = expected_statuses
        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.poll_sleep = poll_sleep
        self.impersonation_chain = impersonation_chain

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize class arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.dataflow.DataflowJobStatusTrigger",
            {
                "job_id": self.job_id,
                "expected_statuses": self.expected_statuses,
                "project_id": self.project_id,
                "location": self.location,
                "gcp_conn_id": self.gcp_conn_id,
                "poll_sleep": self.poll_sleep,
                "impersonation_chain": self.impersonation_chain,
            },
        )

    async def run(self):
        """
        Loop until the job reaches an expected or terminal state.

        Yields a TriggerEvent with success status, if the client returns an expected job status.

        Yields a TriggerEvent with error status, if the client returns an unexpected terminal
        job status or any exception is raised while looping.

        In any other case the Trigger will wait for a specified amount of time
        stored in self.poll_sleep variable.
        """
        try:
            while True:
                job_status = await self.async_hook.get_job_status(
                    job_id=self.job_id,
                    project_id=self.project_id,
                    location=self.location,
                )
                if job_status.name in self.expected_statuses:
                    yield TriggerEvent(
                        {
                            "status": "success",
                            "message": f"Job with id '{self.job_id}' has reached an expected state: {job_status.name}",
                        }
                    )
                    return
                elif job_status.name in DataflowJobStatus.TERMINAL_STATES:
                    yield TriggerEvent(
                        {
                            "status": "error",
                            "message": f"Job with id '{self.job_id}' is already in terminal state: {job_status.name}",
                        }
                    )
                    return
                self.log.info("Sleeping for %s seconds.", self.poll_sleep)
                await asyncio.sleep(self.poll_sleep)
        except Exception as e:
            self.log.error("Exception occurred while checking for job status!")
            yield TriggerEvent(
                {
                    "status": "error",
                    "message": str(e),
                }
            )

    @cached_property
    def async_hook(self) -> AsyncDataflowHook:
        return AsyncDataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            poll_sleep=self.poll_sleep,
            impersonation_chain=self.impersonation_chain,
        )


class DataflowStartYamlJobTrigger(BaseTrigger):
    """
    Dataflow trigger that checks the state of a Dataflow YAML job.

    :param job_id: Required. ID of the job.
    :param project_id: Required. The Google Cloud project ID in which the job was started.
    :param location: The location where job is executed. If set to None then
        the value of DEFAULT_DATAFLOW_LOCATION will be used.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param poll_sleep: Optional. The time in seconds to sleep between polling Google Cloud Platform
        for the Dataflow job.
    :param cancel_timeout: Optional. How long (in seconds) operator should wait for the pipeline to be
        successfully cancelled when task is being killed.
    :param expected_terminal_state: Optional. The expected terminal state of the Dataflow job at which the
        operator task is set to succeed. Defaults to 'JOB_STATE_DONE' for the batch jobs and
        'JOB_STATE_RUNNING' for the streaming jobs.
    :param impersonation_chain: Optional. Service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    def __init__(
        self,
        job_id: str,
        project_id: str | None,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        poll_sleep: int = 10,
        cancel_timeout: int | None = 5 * 60,
        expected_terminal_state: str | None = None,
        impersonation_chain: str | Sequence[str] | None = None,
    ):
        super().__init__()
        self.project_id = project_id
        self.job_id = job_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.poll_sleep = poll_sleep
        self.cancel_timeout = cancel_timeout
        self.expected_terminal_state = expected_terminal_state
        self.impersonation_chain = impersonation_chain

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize class arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.dataflow.DataflowStartYamlJobTrigger",
            {
                "project_id": self.project_id,
                "job_id": self.job_id,
                "location": self.location,
                "gcp_conn_id": self.gcp_conn_id,
                "poll_sleep": self.poll_sleep,
                "expected_terminal_state": self.expected_terminal_state,
                "impersonation_chain": self.impersonation_chain,
                "cancel_timeout": self.cancel_timeout,
            },
        )

    async def run(self):
        """
        Fetch job and yield events depending on the job's type and state.

        Yield TriggerEvent if the job reaches a terminal state.
        Otherwise awaits for a specified amount of time stored in self.poll_sleep variable.
        """
        hook: AsyncDataflowHook = self._get_async_hook()
        try:
            while True:
                job: Job = await hook.get_job(
                    job_id=self.job_id,
                    project_id=self.project_id,
                    location=self.location,
                )
                job_state = job.current_state
                job_type = job.type_
                if job_state.name == self.expected_terminal_state:
                    yield TriggerEvent(
                        {
                            "job": Job.to_dict(job),
                            "status": "success",
                            "message": f"Job reached the expected terminal state: {self.expected_terminal_state}.",
                        }
                    )
                    return
                elif job_type == JobType.JOB_TYPE_STREAMING and job_state == JobState.JOB_STATE_RUNNING:
                    yield TriggerEvent(
                        {
                            "job": Job.to_dict(job),
                            "status": "success",
                            "message": "Streaming job reached the RUNNING state.",
                        }
                    )
                    return
                elif job_type == JobType.JOB_TYPE_BATCH and job_state == JobState.JOB_STATE_DONE:
                    yield TriggerEvent(
                        {
                            "job": Job.to_dict(job),
                            "status": "success",
                            "message": "Batch job completed.",
                        }
                    )
                    return
                elif job_state == JobState.JOB_STATE_FAILED:
                    yield TriggerEvent(
                        {
                            "job": Job.to_dict(job),
                            "status": "error",
                            "message": "Job failed.",
                        }
                    )
                    return
                elif job_state == JobState.JOB_STATE_STOPPED:
                    yield TriggerEvent(
                        {
                            "job": Job.to_dict(job),
                            "status": "stopped",
                            "message": "Job was stopped.",
                        }
                    )
                    return
                else:
                    self.log.info("Current job status is: %s", job_state.name)
                    self.log.info("Sleeping for %s seconds.", self.poll_sleep)
                    await asyncio.sleep(self.poll_sleep)
        except Exception as e:
            self.log.exception("Exception occurred while checking for job completion.")
            yield TriggerEvent({"job": None, "status": "error", "message": str(e)})

    def _get_async_hook(self) -> AsyncDataflowHook:
        return AsyncDataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            poll_sleep=self.poll_sleep,
            impersonation_chain=self.impersonation_chain,
            cancel_timeout=self.cancel_timeout,
        )


class DataflowJobMetricsTrigger(BaseTrigger):
    """
    Trigger that checks for metrics associated with a Dataflow job.

    :param job_id: Required. ID of the job.
    :param project_id: Required. The Google Cloud project ID in which the job was started.
    :param location: Optional. The location where the job is executed. If set to None then
        the value of DEFAULT_DATAFLOW_LOCATION will be used.
    :param gcp_conn_id: The connection ID to use for connecting to Google Cloud.
    :param poll_sleep: Time (seconds) to wait between two consecutive calls to check the job.
    :param impersonation_chain: Optional. Service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param fail_on_terminal_state: If set to True the trigger will yield a TriggerEvent with
        error status if the job reaches a terminal state.
    """

    def __init__(
        self,
        job_id: str,
        project_id: str | None,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        poll_sleep: int = 10,
        impersonation_chain: str | Sequence[str] | None = None,
        fail_on_terminal_state: bool = True,
    ):
        super().__init__()
        self.project_id = project_id
        self.job_id = job_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.poll_sleep = poll_sleep
        self.impersonation_chain = impersonation_chain
        self.fail_on_terminal_state = fail_on_terminal_state

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize class arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.dataflow.DataflowJobMetricsTrigger",
            {
                "project_id": self.project_id,
                "job_id": self.job_id,
                "location": self.location,
                "gcp_conn_id": self.gcp_conn_id,
                "poll_sleep": self.poll_sleep,
                "impersonation_chain": self.impersonation_chain,
                "fail_on_terminal_state": self.fail_on_terminal_state,
            },
        )

    async def run(self):
        """
        Loop until a terminal job status or any job metrics are returned.

        Yields a TriggerEvent with success status, if the client returns any job metrics
        and fail_on_terminal_state attribute is False.

        Yields a TriggerEvent with error status, if the client returns a job status with
        a terminal state value and fail_on_terminal_state attribute is True.

        Yields a TriggerEvent with error status, if any exception is raised while looping.

        In any other case the Trigger will wait for a specified amount of time
        stored in self.poll_sleep variable.
        """
        try:
            while True:
                job_status = await self.async_hook.get_job_status(
                    job_id=self.job_id,
                    project_id=self.project_id,
                    location=self.location,
                )
                job_metrics = await self.get_job_metrics()
                if self.fail_on_terminal_state and job_status.name in DataflowJobStatus.TERMINAL_STATES:
                    yield TriggerEvent(
                        {
                            "status": "error",
                            "message": f"Job with id '{self.job_id}' is already in terminal state: {job_status.name}",
                            "result": None,
                        }
                    )
                    return
                if job_metrics:
                    yield TriggerEvent(
                        {
                            "status": "success",
                            "message": f"Detected {len(job_metrics)} metrics for job '{self.job_id}'",
                            "result": job_metrics,
                        }
                    )
                    return
                self.log.info("Sleeping for %s seconds.", self.poll_sleep)
                await asyncio.sleep(self.poll_sleep)
        except Exception as e:
            self.log.error("Exception occurred while checking for job's metrics!")
            yield TriggerEvent({"status": "error", "message": str(e), "result": None})

    async def get_job_metrics(self) -> list[dict[str, Any]]:
        """Wait for the Dataflow client response and then return it in a serialized list."""
        job_response: JobMetrics = await self.async_hook.get_job_metrics(
            job_id=self.job_id,
            project_id=self.project_id,
            location=self.location,
        )
        return self._get_metrics_from_job_response(job_response)

    def _get_metrics_from_job_response(self, job_response: JobMetrics) -> list[dict[str, Any]]:
        """Return a list of serialized MetricUpdate objects."""
        return [MetricUpdate.to_dict(metric) for metric in job_response.metrics]

    @cached_property
    def async_hook(self) -> AsyncDataflowHook:
        return AsyncDataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            poll_sleep=self.poll_sleep,
            impersonation_chain=self.impersonation_chain,
        )


class DataflowJobAutoScalingEventTrigger(BaseTrigger):
    """
    Trigger that checks for autoscaling events associated with a Dataflow job.

    :param job_id: Required. ID of the job.
    :param project_id: Required. The Google Cloud project ID in which the job was started.
    :param location: Optional. The location where the job is executed. If set to None then
        the value of DEFAULT_DATAFLOW_LOCATION will be used.
    :param gcp_conn_id: The connection ID to use for connecting to Google Cloud.
    :param poll_sleep: Time (seconds) to wait between two consecutive calls to check the job.
    :param impersonation_chain: Optional. Service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param fail_on_terminal_state: If set to True the trigger will yield a TriggerEvent with
        error status if the job reaches a terminal state.
    """

    def __init__(
        self,
        job_id: str,
        project_id: str | None,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        poll_sleep: int = 10,
        impersonation_chain: str | Sequence[str] | None = None,
        fail_on_terminal_state: bool = True,
    ):
        super().__init__()
        self.project_id = project_id
        self.job_id = job_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.poll_sleep = poll_sleep
        self.impersonation_chain = impersonation_chain
        self.fail_on_terminal_state = fail_on_terminal_state

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize class arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.dataflow.DataflowJobAutoScalingEventTrigger",
            {
                "project_id": self.project_id,
                "job_id": self.job_id,
                "location": self.location,
                "gcp_conn_id": self.gcp_conn_id,
                "poll_sleep": self.poll_sleep,
                "impersonation_chain": self.impersonation_chain,
                "fail_on_terminal_state": self.fail_on_terminal_state,
            },
        )

    async def run(self):
        """
        Loop until a terminal job status or any autoscaling events are returned.

        Yields a TriggerEvent with success status, if the client returns any autoscaling events
        and fail_on_terminal_state attribute is False.

        Yields a TriggerEvent with error status, if the client returns a job status with
        a terminal state value and fail_on_terminal_state attribute is True.

        Yields a TriggerEvent with error status, if any exception is raised while looping.

        In any other case the Trigger will wait for a specified amount of time
        stored in self.poll_sleep variable.
        """
        try:
            while True:
                job_status = await self.async_hook.get_job_status(
                    job_id=self.job_id,
                    project_id=self.project_id,
                    location=self.location,
                )
                autoscaling_events = await self.list_job_autoscaling_events()
                if self.fail_on_terminal_state and job_status.name in DataflowJobStatus.TERMINAL_STATES:
                    yield TriggerEvent(
                        {
                            "status": "error",
                            "message": f"Job with id '{self.job_id}' is already in terminal state: {job_status.name}",
                            "result": None,
                        }
                    )
                    return
                if autoscaling_events:
                    yield TriggerEvent(
                        {
                            "status": "success",
                            "message": f"Detected {len(autoscaling_events)} autoscaling events for job '{self.job_id}'",
                            "result": autoscaling_events,
                        }
                    )
                    return
                self.log.info("Sleeping for %s seconds.", self.poll_sleep)
                await asyncio.sleep(self.poll_sleep)
        except Exception as e:
            self.log.error("Exception occurred while checking for job's autoscaling events!")
            yield TriggerEvent({"status": "error", "message": str(e), "result": None})

    async def list_job_autoscaling_events(self) -> list[dict[str, str | dict]]:
        """Wait for the Dataflow client response and then return it in a serialized list."""
        job_response: ListJobMessagesAsyncPager = await self.async_hook.list_job_messages(
            job_id=self.job_id,
            project_id=self.project_id,
            location=self.location,
        )
        return self._get_autoscaling_events_from_job_response(job_response)

    def _get_autoscaling_events_from_job_response(
        self, job_response: ListJobMessagesAsyncPager
    ) -> list[dict[str, str | dict]]:
        """Return a list of serialized AutoscalingEvent objects."""
        return [AutoscalingEvent.to_dict(event) for event in job_response.autoscaling_events]

    @cached_property
    def async_hook(self) -> AsyncDataflowHook:
        return AsyncDataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            poll_sleep=self.poll_sleep,
            impersonation_chain=self.impersonation_chain,
        )


class DataflowJobMessagesTrigger(BaseTrigger):
    """
    Trigger that checks for job messages associated with a Dataflow job.

    :param job_id: Required. ID of the job.
    :param project_id: Required. The Google Cloud project ID in which the job was started.
    :param location: Optional. The location where the job is executed. If set to None then
        the value of DEFAULT_DATAFLOW_LOCATION will be used.
    :param gcp_conn_id: The connection ID to use for connecting to Google Cloud.
    :param poll_sleep: Time (seconds) to wait between two consecutive calls to check the job.
    :param impersonation_chain: Optional. Service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param fail_on_terminal_state: If set to True the trigger will yield a TriggerEvent with
        error status if the job reaches a terminal state.
    """

    def __init__(
        self,
        job_id: str,
        project_id: str | None,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        poll_sleep: int = 10,
        impersonation_chain: str | Sequence[str] | None = None,
        fail_on_terminal_state: bool = True,
    ):
        super().__init__()
        self.project_id = project_id
        self.job_id = job_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.poll_sleep = poll_sleep
        self.impersonation_chain = impersonation_chain
        self.fail_on_terminal_state = fail_on_terminal_state

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize class arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.dataflow.DataflowJobMessagesTrigger",
            {
                "project_id": self.project_id,
                "job_id": self.job_id,
                "location": self.location,
                "gcp_conn_id": self.gcp_conn_id,
                "poll_sleep": self.poll_sleep,
                "impersonation_chain": self.impersonation_chain,
                "fail_on_terminal_state": self.fail_on_terminal_state,
            },
        )

    async def run(self):
        """
        Loop until a terminal job status or any job messages are returned.

        Yields a TriggerEvent with success status, if the client returns any job messages
        and fail_on_terminal_state attribute is False.

        Yields a TriggerEvent with error status, if the client returns a job status with
        a terminal state value and fail_on_terminal_state attribute is True.

        Yields a TriggerEvent with error status, if any exception is raised while looping.

        In any other case the Trigger will wait for a specified amount of time
        stored in self.poll_sleep variable.
        """
        try:
            while True:
                job_status = await self.async_hook.get_job_status(
                    job_id=self.job_id,
                    project_id=self.project_id,
                    location=self.location,
                )
                job_messages = await self.list_job_messages()
                if self.fail_on_terminal_state and job_status.name in DataflowJobStatus.TERMINAL_STATES:
                    yield TriggerEvent(
                        {
                            "status": "error",
                            "message": f"Job with id '{self.job_id}' is already in terminal state: {job_status.name}",
                            "result": None,
                        }
                    )
                    return
                if job_messages:
                    yield TriggerEvent(
                        {
                            "status": "success",
                            "message": f"Detected {len(job_messages)} job messages for job '{self.job_id}'",
                            "result": job_messages,
                        }
                    )
                    return
                self.log.info("Sleeping for %s seconds.", self.poll_sleep)
                await asyncio.sleep(self.poll_sleep)
        except Exception as e:
            self.log.error("Exception occurred while checking for job's messages!")
            yield TriggerEvent({"status": "error", "message": str(e), "result": None})

    async def list_job_messages(self) -> list[dict[str, str | dict]]:
        """Wait for the Dataflow client response and then return it in a serialized list."""
        job_response: ListJobMessagesAsyncPager = await self.async_hook.list_job_messages(
            job_id=self.job_id,
            project_id=self.project_id,
            location=self.location,
        )
        return self._get_job_messages_from_job_response(job_response)

    def _get_job_messages_from_job_response(
        self, job_response: ListJobMessagesAsyncPager
    ) -> list[dict[str, str | dict]]:
        """Return a list of serialized JobMessage objects."""
        return [JobMessage.to_dict(message) for message in job_response.job_messages]

    @cached_property
    def async_hook(self) -> AsyncDataflowHook:
        return AsyncDataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            poll_sleep=self.poll_sleep,
            impersonation_chain=self.impersonation_chain,
        )


class DataflowJobStateCompleteTrigger(BaseTrigger):
    """
    Trigger that monitors if a Dataflow job has reached any of successful terminal state meant for that job.

    :param job_id: Required. ID of the job.
    :param project_id: Required. The Google Cloud project ID in which the job was started.
    :param location: Optional. The location where the job is executed. If set to None then
        the value of DEFAULT_DATAFLOW_LOCATION will be used.
    :param wait_until_finished: Optional. Dataflow option to block pipeline until completion.
    :param gcp_conn_id: The connection ID to use for connecting to Google Cloud.
    :param poll_sleep: Time (seconds) to wait between two consecutive calls to check the job.
    :param impersonation_chain: Optional. Service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    def __init__(
        self,
        job_id: str,
        project_id: str | None,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        wait_until_finished: bool | None = None,
        gcp_conn_id: str = "google_cloud_default",
        poll_sleep: int = 10,
        impersonation_chain: str | Sequence[str] | None = None,
    ):
        super().__init__()
        self.job_id = job_id
        self.project_id = project_id
        self.location = location
        self.wait_until_finished = wait_until_finished
        self.gcp_conn_id = gcp_conn_id
        self.poll_sleep = poll_sleep
        self.impersonation_chain = impersonation_chain

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize class arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.dataflow.DataflowJobStateCompleteTrigger",
            {
                "job_id": self.job_id,
                "project_id": self.project_id,
                "location": self.location,
                "wait_until_finished": self.wait_until_finished,
                "gcp_conn_id": self.gcp_conn_id,
                "poll_sleep": self.poll_sleep,
                "impersonation_chain": self.impersonation_chain,
            },
        )

    async def run(self):
        """
        Loop until the job reaches  successful final or error state.

        Yields a TriggerEvent with success status, if the job reaches successful state for own type.

        Yields a TriggerEvent with error status, if the client returns an unexpected terminal
        job status or any exception is raised while looping.

        In any other case the Trigger will wait for a specified amount of time
        stored in self.poll_sleep variable.
        """
        try:
            while True:
                job = await self.async_hook.get_job(
                    project_id=self.project_id,
                    job_id=self.job_id,
                    location=self.location,
                )
                job_state = job.current_state.name
                job_type_name = job.type_.name

                FAILED_STATES = DataflowJobStatus.FAILED_END_STATES | {DataflowJobStatus.JOB_STATE_DRAINED}
                if job_state in FAILED_STATES:
                    yield TriggerEvent(
                        {
                            "status": "error",
                            "message": (
                                f"Job with id '{self.job_id}' is in failed terminal state: {job_state}"
                            ),
                        }
                    )
                    return

                if self.async_hook.job_reached_terminal_state(
                    job={"id": self.job_id, "currentState": job_state, "type": job_type_name},
                    wait_until_finished=self.wait_until_finished,
                ):
                    yield TriggerEvent(
                        {
                            "status": "success",
                            "message": (
                                f"Job with id '{self.job_id}' has reached successful final state: {job_state}"
                            ),
                        }
                    )
                    return
                self.log.info("Sleeping for %s seconds.", self.poll_sleep)
                await asyncio.sleep(self.poll_sleep)
        except Exception as e:
            self.log.error("Exception occurred while checking for job state!")
            yield TriggerEvent(
                {
                    "status": "error",
                    "message": str(e),
                }
            )

    @cached_property
    def async_hook(self) -> AsyncDataflowHook:
        return AsyncDataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            poll_sleep=self.poll_sleep,
            impersonation_chain=self.impersonation_chain,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from google.api_core.exceptions import GoogleAPICallError
from google.cloud.storage_transfer_v1 import TransferOperation
from google.protobuf import struct_pb2

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.cloud_storage_transfer_service import (
    CloudDataTransferServiceAsyncHook,
    GcpTransferOperationStatus,
)
from airflow.providers.google.cloud.triggers.cloud_storage_transfer_service import (
    CloudDataTransferServiceRunJobTrigger,
    CloudStorageTransferServiceCheckJobStatusTrigger,
    CloudStorageTransferServiceCreateJobsTrigger,
)
from airflow.triggers.base import TriggerEvent

PROJECT_ID = "test-project"
GCP_CONN_ID = "google-cloud-default-id"
JOB_0 = "test-job-0"
JOB_1 = "test-job-1"
JOB_NAMES = [JOB_0, JOB_1]
LATEST_OPERATION_NAME_0 = "test-latest-operation-0"
LATEST_OPERATION_NAME_1 = "test-latest-operation-1"
LATEST_OPERATION_NAMES = [LATEST_OPERATION_NAME_0, LATEST_OPERATION_NAME_1]
POLL_INTERVAL = 2
CLASS_PATH = (
    "airflow.providers.google.cloud.triggers.cloud_storage_transfer_service"
    ".CloudStorageTransferServiceCreateJobsTrigger"
)
ASYNC_HOOK_CLASS_PATH = (
    "airflow.providers.google.cloud.hooks.cloud_storage_transfer_service.CloudDataTransferServiceAsyncHook"
)
EXPECTED_STATUSES = GcpTransferOperationStatus.SUCCESS
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]


@pytest.fixture(scope="session")
def trigger():
    return CloudStorageTransferServiceCreateJobsTrigger(
        project_id=PROJECT_ID,
        job_names=JOB_NAMES,
        poll_interval=POLL_INTERVAL,
        gcp_conn_id=GCP_CONN_ID,
    )


def mock_jobs(names: list[str], latest_operation_names: list[str | None]):
    """Returns object that mocks asynchronous looping over mock jobs"""
    jobs = [mock.MagicMock(latest_operation_name=name) for name in latest_operation_names]
    for job, name in zip(jobs, names):
        job.name = name
    mock_obj = mock.MagicMock()
    mock_obj.__aiter__.return_value = iter(jobs)
    return mock_obj


def create_mock_operation(status: TransferOperation.Status, name: str) -> mock.MagicMock:
    _obj = mock.MagicMock(status=status)
    _obj.name = name
    return _obj


class TestCloudStorageTransferServiceCreateJobsTrigger:
    def test_serialize(self, trigger):
        class_path, serialized = trigger.serialize()

        assert class_path == CLASS_PATH
        assert serialized == {
            "project_id": PROJECT_ID,
            "job_names": JOB_NAMES,
            "poll_interval": POLL_INTERVAL,
            "gcp_conn_id": GCP_CONN_ID,
        }

    def test_get_async_hook(self, trigger):
        hook = trigger.get_async_hook()

        assert isinstance(hook, CloudDataTransferServiceAsyncHook)
        assert hook.project_id == PROJECT_ID

    @pytest.mark.asyncio
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_latest_operation")
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_jobs")
    async def test_run(self, get_jobs, get_latest_operation, trigger):
        get_jobs.return_value = mock_jobs(names=JOB_NAMES, latest_operation_names=LATEST_OPERATION_NAMES)
        get_latest_operation.side_effect = [
            create_mock_operation(status=TransferOperation.Status.SUCCESS, name="operation_" + job_name)
            for job_name in JOB_NAMES
        ]
        expected_event = TriggerEvent(
            {
                "status": "success",
                "message": f"Transfer jobs {JOB_0}, {JOB_1} completed successfully",
            }
        )
        generator = trigger.run()
        actual_event = await generator.asend(None)

        assert actual_event == expected_event

    @pytest.mark.parametrize(
        "status",
        [
            TransferOperation.Status.STATUS_UNSPECIFIED,
            TransferOperation.Status.IN_PROGRESS,
            TransferOperation.Status.PAUSED,
            TransferOperation.Status.QUEUED,
        ],
    )
    @pytest.mark.asyncio
    @mock.patch("asyncio.sleep")
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_latest_operation", autospec=True)
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_jobs", autospec=True)
    async def test_run_poll_interval(self, get_jobs, get_latest_operation, mock_sleep, trigger, status):
        get_jobs.side_effect = [
            mock_jobs(names=JOB_NAMES, latest_operation_names=LATEST_OPERATION_NAMES),
            mock_jobs(names=JOB_NAMES, latest_operation_names=LATEST_OPERATION_NAMES),
        ]
        get_latest_operation.side_effect = [
            create_mock_operation(status=status, name="operation_" + job_name) for job_name in JOB_NAMES
        ] + [
            create_mock_operation(status=TransferOperation.Status.SUCCESS, name="operation_" + job_name)
            for job_name in JOB_NAMES
        ]
        expected_event = TriggerEvent(
            {
                "status": "success",
                "message": f"Transfer jobs {JOB_0}, {JOB_1} completed successfully",
            }
        )

        generator = trigger.run()
        actual_event = await generator.asend(None)

        assert actual_event == expected_event
        mock_sleep.assert_called_once_with(POLL_INTERVAL)

    @pytest.mark.parametrize(
        ("latest_operations_names", "expected_failed_job"),
        [
            ([None, LATEST_OPERATION_NAME_1], JOB_0),
            ([LATEST_OPERATION_NAME_0, None], JOB_1),
        ],
    )
    @pytest.mark.asyncio
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_latest_operation")
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_jobs")
    async def test_run_error_job_has_no_latest_operation(
        self, get_jobs, get_latest_operation, trigger, latest_operations_names, expected_failed_job
    ):
        get_jobs.return_value = mock_jobs(names=JOB_NAMES, latest_operation_names=latest_operations_names)
        get_latest_operation.side_effect = [
            create_mock_operation(status=TransferOperation.Status.SUCCESS, name="operation_" + job_name)
            if job_name
            else None
            for job_name in latest_operations_names
        ]
        expected_event = TriggerEvent(
            {
                "status": "error",
                "message": f"Transfer job {expected_failed_job} has no latest operation.",
            }
        )

        generator = trigger.run()
        actual_event = await generator.asend(None)

        assert actual_event == expected_event

    @pytest.mark.parametrize(
        ("job_statuses", "failed_operation", "expected_status"),
        [
            (
                [TransferOperation.Status.ABORTED, TransferOperation.Status.SUCCESS],
                LATEST_OPERATION_NAME_0,
                "ABORTED",
            ),
            (
                [TransferOperation.Status.FAILED, TransferOperation.Status.SUCCESS],
                LATEST_OPERATION_NAME_0,
                "FAILED",
            ),
            (
                [TransferOperation.Status.SUCCESS, TransferOperation.Status.ABORTED],
                LATEST_OPERATION_NAME_1,
                "ABORTED",
            ),
            (
                [TransferOperation.Status.SUCCESS, TransferOperation.Status.FAILED],
                LATEST_OPERATION_NAME_1,
                "FAILED",
            ),
        ],
    )
    @pytest.mark.asyncio
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_latest_operation")
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_jobs")
    async def test_run_error_one_job_failed_or_aborted(
        self,
        get_jobs,
        get_latest_operation,
        trigger,
        job_statuses,
        failed_operation,
        expected_status,
    ):
        get_jobs.return_value = mock_jobs(names=JOB_NAMES, latest_operation_names=LATEST_OPERATION_NAMES)
        get_latest_operation.side_effect = [
            create_mock_operation(status=status, name=operation_name)
            for status, operation_name in zip(job_statuses, LATEST_OPERATION_NAMES)
        ]
        expected_event = TriggerEvent(
            {
                "status": "error",
                "message": f"Transfer operation {failed_operation} failed with status {expected_status}",
            }
        )

        generator = trigger.run()
        actual_event = await generator.asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_latest_operation")
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_jobs")
    async def test_run_get_jobs_airflow_exception(self, get_jobs, get_latest_operation, trigger):
        expected_error_message = "Mock error message"
        get_jobs.side_effect = AirflowException(expected_error_message)

        get_latest_operation.side_effect = [
            create_mock_operation(status=TransferOperation.Status.SUCCESS, name="operation_" + job_name)
            for job_name in JOB_NAMES
        ]
        expected_event = TriggerEvent(
            {
                "status": "error",
                "message": expected_error_message,
            }
        )

        generator = trigger.run()
        actual_event = await generator.asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_latest_operation")
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_jobs")
    async def test_run_get_latest_operation_airflow_exception(self, get_jobs, get_latest_operation, trigger):
        get_jobs.return_value = mock_jobs(names=JOB_NAMES, latest_operation_names=LATEST_OPERATION_NAMES)
        expected_error_message = "Mock error message"
        get_latest_operation.side_effect = AirflowException(expected_error_message)

        expected_event = TriggerEvent(
            {
                "status": "error",
                "message": expected_error_message,
            }
        )

        generator = trigger.run()
        actual_event = await generator.asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_latest_operation")
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".get_jobs")
    async def test_run_get_latest_operation_google_api_call_error(
        self, get_jobs, get_latest_operation, trigger
    ):
        get_jobs.return_value = mock_jobs(names=JOB_NAMES, latest_operation_names=LATEST_OPERATION_NAMES)
        error_message = "Mock error message"
        get_latest_operation.side_effect = GoogleAPICallError(error_message)

        expected_event = TriggerEvent(
            {
                "status": "error",
                "message": f"{None} {error_message}",
            }
        )

        generator = trigger.run()
        actual_event = await generator.asend(None)

        assert actual_event == expected_event


class TestCloudStorageTransferServiceCheckJobStatusTrigger:
    @pytest.fixture
    def trigger(self):
        return CloudStorageTransferServiceCheckJobStatusTrigger(
            project_id=PROJECT_ID,
            job_name=JOB_0,
            expected_statuses=EXPECTED_STATUSES,
            poke_interval=POLL_INTERVAL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

    def test_serialize(self, trigger):
        class_path, serialized = trigger.serialize()
        assert class_path == (
            "airflow.providers.google.cloud.triggers.cloud_storage_transfer_service"
            ".CloudStorageTransferServiceCheckJobStatusTrigger"
        )
        assert serialized == {
            "project_id": PROJECT_ID,
            "job_name": JOB_0,
            "expected_statuses": EXPECTED_STATUSES,
            "poke_interval": POLL_INTERVAL,
            "gcp_conn_id": GCP_CONN_ID,
            "impersonation_chain": IMPERSONATION_CHAIN,
        }

    @pytest.mark.parametrize(
        ("attr", "expected_value"),
        [
            ("gcp_conn_id", GCP_CONN_ID),
            ("impersonation_chain", IMPERSONATION_CHAIN),
        ],
    )
    def test_get_async_hook(self, attr, expected_value, trigger):
        hook = trigger._get_async_hook()
        actual_value = hook._hook_kwargs.get(attr)
        assert isinstance(hook, CloudDataTransferServiceAsyncHook)
        assert hook._hook_kwargs is not None
        assert actual_value == expected_value

    @pytest.mark.asyncio
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".list_transfer_operations")
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".operations_contain_expected_statuses")
    async def test_run_returns_success_event(
        self,
        operations_contain_expected_statuses,
        list_transfer_operations,
        trigger,
    ):
        operations_contain_expected_statuses.side_effect = [
            False,
            True,
        ]
        expected_event = TriggerEvent(
            {
                "status": "success",
                "message": "Transfer operation completed successfully",
                "operations": list_transfer_operations.return_value,
            }
        )

        actual_event = await trigger.run().asend(None)

        assert actual_event == expected_event
        assert operations_contain_expected_statuses.call_count == 2

    @pytest.mark.asyncio
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".list_transfer_operations")
    async def test_run_returns_exception_event(
        self,
        list_transfer_operations,
        trigger,
    ):
        list_transfer_operations.side_effect = Exception("Transfer operation failed")
        expected_event = TriggerEvent(
            {
                "status": "error",
                "message": "Transfer operation failed",
            }
        )

        actual_event = await trigger.run().asend(None)

        assert actual_event == expected_event


class TestCloudDataTransferServiceRunJobTrigger:
    @pytest.fixture
    def trigger(self):
        return CloudDataTransferServiceRunJobTrigger(
            project_id=PROJECT_ID,
            job_name=JOB_0,
            poke_interval=POLL_INTERVAL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

    def test_serialize(self, trigger):
        class_path, serialized = trigger.serialize()
        assert class_path == (
            "airflow.providers.google.cloud.triggers.cloud_storage_transfer_service"
            ".CloudDataTransferServiceRunJobTrigger"
        )
        assert serialized == {
            "project_id": PROJECT_ID,
            "job_name": JOB_0,
            "poke_interval": POLL_INTERVAL,
            "gcp_conn_id": GCP_CONN_ID,
            "impersonation_chain": IMPERSONATION_CHAIN,
        }

    @pytest.mark.parametrize(
        ("attr", "expected_value"),
        [
            ("gcp_conn_id", GCP_CONN_ID),
            ("impersonation_chain", IMPERSONATION_CHAIN),
        ],
    )
    def test_get_async_hook(self, attr, expected_value, trigger):
        hook = trigger._get_async_hook()
        actual_value = hook._hook_kwargs.get(attr)
        assert isinstance(hook, CloudDataTransferServiceAsyncHook)
        assert hook._hook_kwargs is not None
        assert actual_value == expected_value

    @pytest.mark.asyncio
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".run_transfer_job")
    async def test_run_returns_success_event(
        self,
        run_transfer_job,
        trigger,
    ):
        test_metadata = struct_pb2.Struct()
        test_metadata.update({"test": "test_metadata"})
        test_response = struct_pb2.Struct()
        test_response.update({"test": "test_response"})
        test_operation = mock.Mock(metadata=test_metadata, response=test_response)
        test_operation.name = "test_name"
        run_transfer_job.return_value.operation = test_operation
        run_transfer_job.done.side_effect = True
        expected_event = TriggerEvent(
            {
                "status": "success",
                "message": "Transfer operation run completed successfully",
                "job_result": {
                    "name": "test_name",
                    "metadata": {"test": "test_metadata"},
                    "response": {"test": "test_response"},
                },
            }
        )

        actual_event = await trigger.run().asend(None)

        assert actual_event == expected_event
        assert run_transfer_job.call_count == 1

    @pytest.mark.asyncio
    @mock.patch(ASYNC_HOOK_CLASS_PATH + ".run_transfer_job")
    async def test_run_returns_exception_event(
        self,
        run_transfer_job,
        trigger,
    ):
        run_transfer_job.side_effect = Exception("Run transfer job operation failed")
        expected_event = TriggerEvent(
            {
                "status": "error",
                "message": "Run transfer job operation failed",
            }
        )

        actual_event = await trigger.run().asend(None)

        assert actual_event == expected_event



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for Google BigQuery service.
"""

from __future__ import annotations

import os
from datetime import datetime
from pathlib import Path

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.bigquery import (
    BigQueryCreateEmptyDatasetOperator,
    BigQueryCreateTableOperator,
    BigQueryDeleteDatasetOperator,
    BigQueryInsertJobOperator,
)
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
from system.openlineage.operator import OpenLineageTestOperator

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
LOCATION = "us-east1"

TABLE_1 = "table1"
TABLE_2 = "table2"
TABLE_3 = "table3"
TABLE_4 = "table4"

SCHEMA = [
    {"name": "value", "type": "INTEGER", "mode": "REQUIRED"},
    {"name": "name", "type": "STRING", "mode": "NULLABLE"},
    {"name": "ds", "type": "DATE", "mode": "NULLABLE"},
]

DAG_ID = "bigquery_jobs"
DATASET_NAME = f"dataset_{DAG_ID}_{ENV_ID}"
BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}"
DATASET = f"{DATASET_NAME}1"
INSERT_DATE = "2030-01-01"
INSERT_ROWS_QUERY = (
    f"INSERT {DATASET}.{TABLE_1} VALUES "
    f"(42, 'monty python', '{INSERT_DATE}'), "
    f"(42, 'fishy fish', '{INSERT_DATE}');"
)

CTAS_QUERY = (
    f"CREATE OR REPLACE TABLE {DATASET}.{TABLE_3} AS SELECT value, name, ds FROM {DATASET}.{TABLE_1};"
)

with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "bigquery"],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket", bucket_name=BUCKET_NAME, project_id=PROJECT_ID, storage_class="STANDARD"
    )

    create_dataset = BigQueryCreateEmptyDatasetOperator(
        task_id="create_dataset",
        dataset_id=DATASET,
    )

    create_table = BigQueryCreateTableOperator(
        task_id="create_table",
        dataset_id=DATASET,
        table_id=TABLE_1,
        table_resource={
            "schema": {"fields": SCHEMA},
        },
    )

    insert_query_job = BigQueryInsertJobOperator(
        task_id="insert_query_job",
        configuration={
            "query": {
                "query": INSERT_ROWS_QUERY,
                "useLegacySql": False,
                "priority": "BATCH",
            }
        },
    )

    execute_copy_job = BigQueryInsertJobOperator(
        task_id="execute_copy_job",
        configuration={
            "copy": {
                "sourceTables": [{"projectId": PROJECT_ID, "datasetId": DATASET, "tableId": TABLE_1}],
                "destinationTable": {"projectId": PROJECT_ID, "datasetId": DATASET, "tableId": TABLE_2},
                "writeDisposition": "WRITE_TRUNCATE",
                "operationType": "COPY",
            }
        },
    )

    execute_ctas_query = BigQueryInsertJobOperator(
        task_id="execute_ctas_query",
        configuration={
            "query": {
                "query": CTAS_QUERY,
                "useLegacySql": False,
            }
        },
    )

    execute_load_job = BigQueryInsertJobOperator(
        task_id="execute_load_job",
        configuration={
            "load": {
                "sourceUris": ["gs://cloud-samples-data/bigquery/us-states/us-states.csv"],
                "destinationTable": {
                    "projectId": PROJECT_ID,
                    "datasetId": DATASET,
                    "tableId": TABLE_4,
                },
                "schema": {
                    "fields": [
                        {"name": "name", "type": "STRING", "mode": "NULLABLE"},
                        {"name": "post_abbr", "type": "STRING", "mode": "NULLABLE"},
                    ]
                },
                "write_disposition": "WRITE_TRUNCATE",
                "sourceFormat": "CSV",
            }
        },
    )

    execute_extract_job = BigQueryInsertJobOperator(
        task_id="execute_extract_job",
        configuration={
            "extract": {
                "sourceTable": {"projectId": PROJECT_ID, "datasetId": DATASET, "tableId": TABLE_4},
                "destinationUris": [
                    f"gs://{BUCKET_NAME}/extract.csv",
                ],
                "destinationFormat": "CSV",
            }
        },
    )

    delete_dataset = BigQueryDeleteDatasetOperator(
        task_id="delete_dataset",
        dataset_id=DATASET,
        delete_contents=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    # if not location:
    check_openlineage_events = OpenLineageTestOperator(
        task_id="check_openlineage_events",
        file_path=str(Path(__file__).parent / "resources" / "openlineage" / "bigquery_jobs.json"),
    )
    delete_dataset >> check_openlineage_events

    # TEST SETUP
    create_dataset >> create_table
    # TEST BODY
    create_table >> insert_query_job >> execute_copy_job >> execute_ctas_query >> delete_dataset
    create_dataset >> execute_load_job >> execute_extract_job >> delete_dataset
    # TEST TEARDOWN
    delete_dataset >> delete_bucket

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)
